/* $Id: chpwdfile.c 1.19 Tue, 14 Oct 2003 20:25:02 +0200 warwick $ */

/*
 * Program to manage username:password pairs in an /etc/passwd type
 * authentication file.  The executable will probably have to be setuid.
 */

/*
 * Copyright (c) 2000,2001 Warwick Duncan <warwick@chemeng.uct.ac.za>
 *
 * This software is released under the terms of the GNU GPL v2. See the file
 * LICENSE included with this archive for more information.
 *
 * THIS SOFTWARE IS PROVIDED BY ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <syslog.h>
#include "chpwdfile.h"
#include "util.h"

#ifdef CGI
#include <cgi.h>
#endif	/* CGI */

/********************
 * Global variables *
 ********************/

/* $Format: "static char version[] = \"$ProjectVersion: 0.17 $\";"$ */
static char version[] = "0.24";

/* User details */
struct entry {
    char name[MAXUSERLENGTH+1];
    char passwd[MAXPASSWORDLENGTH+1];
} userinfo;

/* Passwd file details */
struct pwdfile {
    char name[MAXFILENAMELENGTH+1];
    char *name_ptr;
    FILE *stream;
    int fd;	/* file descriptor */
    struct stat pwdstat;
    char tmpname[MAXFILENAMELENGTH+18];	/* temporary file */
    FILE *tmpstream;
    int tmpfd;
    int hash_type;	/* 'd' for des or 'm' for md5 */
    char superpass[MAXPASSWORDLENGTH];	/* admin password */
} fileinfo;

/* The place where we can read some random stuff */
FILE *random_source;

#ifdef CGI
s_cgi *cgi;
#endif

/* Read password directly from stdin (for use in scripts)? */
int passwd_from_stdin=0;

/*************************
 * Start the code proper *
 *************************/

/*
 * Read in a couple of passwords.
 * Return 1 on error, 0 on success.
 * This gets called from within util.c.
 */
int read_new_passwds(char *newpasswd1, char *newpasswd2, int maxlength)
{
#ifdef CGI
    char *passwd;	/* The one we're reading in */

    if ((passwd = cgiGetValue(cgi, CGINEWPASS1)) != NULL) {
	strncpy(newpasswd1, passwd, maxlength-1);
    } else {
	return(1);
    }
    if ((passwd = cgiGetValue(cgi, CGINEWPASS2)) != NULL) {
	strncpy(newpasswd2, passwd, maxlength-1);
    } else {
	return(1);
    }
#else
    if (passwd_from_stdin) {
        fgets(newpasswd1, maxlength-1, stdin);
	 newpasswd1[maxlength] = 0;
        if (strlen(newpasswd1) > 0) newpasswd1[strlen(newpasswd1)-1] = 0;
        strcpy(newpasswd2, newpasswd1);
    }
    else {
        strncpy(newpasswd1, getpass("Enter new password: "), maxlength-1);
        strncpy(newpasswd2, getpass("Confirm new password: "), maxlength-1);
    }
#endif /* CGI */

    return(0);
}

/*
 * Search for user in passwd file; return 1 if found.
 */
int user_search(char *file, char *user)
{
    FILE *stream;
    char line[MAXLINELENGTH], *line_ptr =line;
    char *read_user, *read_passwd;
    int user_found = 0;

    strcpy(fileinfo.superpass, "");

    stream = fopen(file, "r");

    /* Read a line at a time, comparing first field */
    while(fgets(line, sizeof(line), stream) != NULL) {
	line_ptr = line;
	read_user = strsep(&line_ptr, ":");
	read_passwd = strsep(&line_ptr, ":");
	if (read_user == NULL || read_passwd == NULL)
	    return(0);
	if (strncmp(read_user, user, MAXUSERLENGTH) == 0) {
	    strncpy(userinfo.passwd, strsep(&read_passwd, "\n"),
		    sizeof(userinfo.passwd));
	    user_found = 1;
	} else if (strncmp(read_user, ADMIN, MAXUSERLENGTH) == 0) {
	    strncpy(fileinfo.superpass, strsep(&read_passwd, "\n"),
		    sizeof(fileinfo.superpass));
	}
	if (user_found && (strncmp(fileinfo.superpass, "", 
			sizeof(fileinfo.superpass))!=0))
	    break;	/* stop searching the file */
    }

    fclose(stream);

    return(user_found);
}

/*
 * Function to add a user's entry.
 */
int add_user(int minpasswd, int authenticate)
{
    int user_found = 0;	/* we haven't found him yet */
    char *read_passwd;

    /* Look for user in existing passwd file */
    user_found = user_search(fileinfo.name, userinfo.name);

    /* Append entry if not already there */
    if (user_found) {
	return(USEREXISTS);
    } else {
	if (authenticate) {
	    /* Authenticate admin password */
#ifdef CGI
	    read_passwd = cgiGetValue(cgi, CGIOLDPASS);
	    if (read_passwd == NULL)
		return(NOAUTHENTICATE);
#else
	    read_passwd = getpass("Enter password to authenticate: ");
#endif	/* CGI */
	    if (!authenticate_user(ADMIN, fileinfo.superpass,
			read_passwd, MAXPASSWORDLENGTH))
		return(NOAUTHENTICATE);
	}
	/* Read in new user password */
	if (get_passwd(userinfo.passwd, sizeof(userinfo.passwd), minpasswd, 
		    fileinfo.hash_type, random_source) != 0) {
	    return(BADPASSWORD);
	}
	/* Append entry to passwd file */
	fileinfo.fd = open(fileinfo.name, O_WRONLY);
	fileinfo.stream = fdopen(fileinfo.fd, "a");
	if (lock_pwdfile(fileinfo.fd, MAXLOCKTRIES) != 0) {
	    return(BADLOCK);
	}
	fprintf(fileinfo.stream, "%s:%s\n", userinfo.name, userinfo.passwd);
	fclose(fileinfo.stream);
    }

    return(0);
}

/*
 * Function to delete a user's entry
 */
int delete_user(int authenticate)
{
    int user_found = 0;	/* we haven't found him yet */
    char line[MAXLINELENGTH], *line_ptr = line;
    char *read_user, *read_passwd;

    /* Look for user in existing passwd file */
    user_found = user_search(fileinfo.name, userinfo.name);

    if (!user_found) {	/* who??? */
	return(USERNOTFOUND);
    } else {
	if (authenticate) {
	    /* Ask for user or admin password */
#ifdef CGI
	    read_passwd = cgiGetValue(cgi, CGIOLDPASS);
	    if (read_passwd == NULL)
		return(NOAUTHENTICATE);
#else
	    read_passwd = getpass("Enter password to authenticate: ");
#endif	/* CGI */
	    if (!authenticate_user(ADMIN, fileinfo.superpass, read_passwd, 
			MAXPASSWORDLENGTH))
		return(NOAUTHENTICATE);
	}
	/* Start modifying the files */
	fileinfo.fd = open(fileinfo.name, O_RDWR);
	fileinfo.stream = fopen(fileinfo.name, "r+");
	if (lock_pwdfile(fileinfo.fd, MAXLOCKTRIES) != 0) {
	    return(BADLOCK);
	}
	if (open_tmpfile(fileinfo.name, fileinfo.tmpname, &fileinfo.tmpfd,
		    &fileinfo.tmpstream) != 0) { 
	    return(BADTMP); 
	}
	/* Read a line at a time, write to tmp file if no match */
	while(fgets(line, sizeof(line), fileinfo.stream) != NULL) {
	    line_ptr = line;
	    read_user = strsep(&line_ptr, ":");
	    read_passwd = line_ptr;
	    if (read_user == NULL || read_passwd == NULL)
		return(0);
	    if (strncmp(read_user, userinfo.name, MAXUSERLENGTH) != 0) {
		fprintf(fileinfo.tmpstream, "%s:%s", read_user, read_passwd);
	    }
	}
	/* Move the tmp file over the original */
	fclose(fileinfo.stream);
	fclose(fileinfo.tmpstream);
	if (geteuid() == 0) {	/* Only root can change file ownership */
	    if ((chown(fileinfo.tmpname, fileinfo.pwdstat.st_uid, 
			    fileinfo.pwdstat.st_gid)) == -1)
		return(FILEPERMERR);
	} else {	/* Set the group ownership only */
	    if ((chown(fileinfo.tmpname, -1, fileinfo.pwdstat.st_gid)) == -1)
		return(FILEPERMERR);
	}
	if ((chmod(fileinfo.tmpname, fileinfo.pwdstat.st_mode)) == -1)
	    return(FILEPERMERR);
	if ((rename(fileinfo.tmpname, fileinfo.name)) == -1)
	    return(FILEPERMERR);
    }

    return(0);
}

/*
 * Function to modify a user's entry
 */
int modify_user(int minpasswd, int authenticate)
{
    int user_found=0;	/* we haven't found him yet */
    char line[MAXLINELENGTH], *line_ptr=line;
    char *read_user, *read_passwd;

    /* Look for user in existing passwd file */
    user_found = user_search(fileinfo.name, userinfo.name);

    if (!user_found) {	/* who??? */
	return(USERNOTFOUND);
    } else {
	if (authenticate) {
	    /* Ask for user or admin password */
#ifdef CGI
	    read_passwd = cgiGetValue(cgi, CGIOLDPASS);
	    if (read_passwd == NULL)
		return(NOAUTHENTICATE);
#else
	    read_passwd = getpass("Enter password to authenticate: ");
#endif	/* CGI */
	    if (!authenticate_user(ADMIN, fileinfo.superpass, read_passwd, 
			MAXPASSWORDLENGTH) 
		    && !authenticate_user(userinfo.name,
			userinfo.passwd, read_passwd, MAXPASSWORDLENGTH))
		return(NOAUTHENTICATE);
	}
	/* Read in new user password */
	if (get_passwd(userinfo.passwd, sizeof(userinfo.passwd),
		    minpasswd, fileinfo.hash_type, random_source) != 0) {
	    return(BADPASSWORD);
	}
	/* Start modifying the files */
	fileinfo.fd = open(fileinfo.name, O_RDWR);
	fileinfo.stream = fopen(fileinfo.name, "r+");
	if (lock_pwdfile(fileinfo.fd, MAXLOCKTRIES) != 0) {
	    return(BADLOCK);
	}
	if (open_tmpfile(fileinfo.name, fileinfo.tmpname, &fileinfo.tmpfd,
		    &fileinfo.tmpstream) != 0) { 
	    return(BADTMP); 
	}
	/* Read a line at a time, doing the right thing for a match */
	while(fgets(line, sizeof(line), fileinfo.stream) != NULL) {
	    line_ptr = line;
	    read_user = strsep(&line_ptr, ":");
	    read_passwd = line_ptr;
	    if (read_user == NULL || read_passwd == NULL)
		return(0);
	    if (strncmp(read_user, userinfo.name, MAXUSERLENGTH) != 0) {
		fprintf(fileinfo.tmpstream, "%s:%s", read_user, read_passwd);
	    } else {
		fprintf(fileinfo.tmpstream, "%s:%s\n", userinfo.name,
			userinfo.passwd);
	    }
	}
	/* Move the tmp file over the original */
	fclose(fileinfo.stream);
	fclose(fileinfo.tmpstream);
	if (geteuid() == 0) {	/* Only root can change file ownership */
	    if ((chown(fileinfo.tmpname, fileinfo.pwdstat.st_uid, 
			    fileinfo.pwdstat.st_gid)) == -1)
		return(FILEPERMERR);
	} else {	/* Set the group ownership only */
	    if ((chown(fileinfo.tmpname, -1, fileinfo.pwdstat.st_gid)) == -1)
		return(FILEPERMERR);
	}
	if ((chmod(fileinfo.tmpname, fileinfo.pwdstat.st_mode)) == -1)
	    return(FILEPERMERR);
	if ((rename(fileinfo.tmpname, fileinfo.name)) == -1)
	    return(FILEPERMERR);
    }

    return(0);
}

/*
 * Read configuration from config file.  Set passwd file to first entry 
 * if not defined yet.
 * Format is keyword=value with # to start comments.
 */
int read_config(FILE *confstream, int *minpasswd, int *minuser)
{
    char line[MAXLINELENGTH+1];
    char linecp[MAXLINELENGTH+1];
    char *linecp_ptr;
    char *keyword, *value;
    struct {
	char pwdfile[MAXFILENAMELENGTH];
	int minpassword;
	int minusername;
	int hashtype;
    } config_entry;

    /* Some things should start undefined */
    config_entry.minpassword = -1;
    config_entry.minusername = -1;
    config_entry.hashtype = -1;

    /* Start looking through the file */
    while (fgets(line, sizeof(line), confstream) != NULL) {
	/* Remove whitespace and comments */
	squash_line(&line[0], &linecp[0]);
	/* Ignore empty lines and comments */
	if (strncmp(linecp, "", sizeof(linecp)) == 0)
	    continue;

	/* Check for entry; must start with "file" */
	linecp_ptr = &linecp[0];
	keyword = strsep(&linecp_ptr, "=");
	value = linecp_ptr;
	if (strncmp(keyword, "file", sizeof(linecp)) != 0)
	    return(BADCONFFILE);
	strncpy(config_entry.pwdfile, value, sizeof(config_entry.pwdfile));

	/* Set passwd file name if not already set */
	if (fileinfo.name_ptr == NULL) {
	    strncpy(fileinfo.name, config_entry.pwdfile, sizeof(fileinfo.name));
	    fileinfo.name_ptr = fileinfo.name;
	}

	while (strncmp(keyword, "end", sizeof(linecp)) != 0) {
	    if (fgets(line, sizeof(linecp), confstream) == NULL)
		return(BADCONFFILE);
	    squash_line(&line[0], &linecp[0]);
	    if (strncmp(linecp, "", sizeof(linecp)) == 0)
		continue;
	    linecp_ptr = &linecp[0];
	    keyword = strsep(&linecp_ptr, "=");
	    value = linecp_ptr;

	    /* Check which option we're setting */
	    if (strncmp(keyword, "minusername", sizeof(linecp)) == 0) {
		config_entry.minusername = atoi(value);
	    } else if (strncmp(keyword, "minpassword", sizeof(linecp)) == 0) {
		config_entry.minpassword = atoi(value);
	    } else if (strncmp(keyword, "hashtype", sizeof(linecp)) == 0) {
		if (*value == 'd' || *value == 'm') {
		    config_entry.hashtype = *value;
		} else {
		    return(BADCONFFILE);
		}
	    } 
	}

	/* Modify the config if this is the file we want */
	if (strncmp(fileinfo.name, config_entry.pwdfile, 
		    sizeof(fileinfo.name)) == 0) {
	    if (config_entry.minpassword > 0)
		*minpasswd = config_entry.minpassword;
	    if (config_entry.minusername > 0)
		*minuser = config_entry.minusername;
	    if (config_entry.hashtype > 0)
		fileinfo.hash_type = config_entry.hashtype;
	    break;	/* stop reading the config file */
	} 
    }

    return(0);
}

/*
 * Print how to use this here program
 */
void print_usage(char *basename, int err)
{
#ifdef CGI
    char **cgi_vars;
#endif /* CGI */

    openlog(basename, 0, LOG_USER);
#ifdef CGI
    cgiHeader();
    puts("<html><head><title>Error</title></head>");
    puts("<body><h1>An error has occurred</h1><p>");
    /* Print message depending on value of err */
    switch (err) {
	case BADHASHTYPE:
	    printf("Bad hashing algorithm (must be 'd' or 'm').\n");
	    syslog(LOG_ERR, "Bad hashing algorithm (must be 'd' or 'm').\n");
	    break;
	case MINPASSWORDTOOBIG:
	    printf("Specified minimum password length is too long.\n");
	    syslog(LOG_ERR, "Specified minimum password length is too long.\n");
	    break;
	case MINUSERTOOBIG:
	    printf("Specified minimum username length is too long.\n");
	    syslog(LOG_ERR, "Specified minimum username length is too long.\n");
	    break;
	case BADCALL:
	    printf("Unknown option.\n");
	    syslog(LOG_ERR, "Unknown option.\n");
	    break;
	case NOCONFFILE:
	    printf("Error opening config file.\n");
	    syslog(LOG_ERR, "Error opening config file.\n");
	    break;
	case BADCONFFILE:
	    printf("Syntax error in config file.\n");
	    syslog(LOG_ERR, "Syntax error in config file.\n");
	    break;
	case USERNOTFOUND:
	    printf("User %s does not exist.\n", userinfo.name);
	    syslog(LOG_ERR, "User %s does not exist in %s.\n", 
		    userinfo.name, fileinfo.name);
	    break;
	case USEREXISTS:
	    printf("User %s already exists.\n", userinfo.name);
	    syslog(LOG_ERR, "User %s already exists in %s.\n",
		    userinfo.name, fileinfo.name);
	    break;
	case NOAUTHENTICATE:
	    printf("Permission to change password file denied.\n");
	    syslog(LOG_ERR, "Permission to change password file denied (%s).\n",
		    userinfo.name);
	    break;
	case BADPASSWORD:
	    printf("Bad password or passwords don't match.\n");
	    syslog(LOG_ERR, "Bad password or passwords don't match (%s).\n",
		    userinfo.name);
	    break;
	case BADLOCK:
	    printf("Could not lock the passwd file.\n");
	    syslog(LOG_ERR, "Could not lock %s.\n", fileinfo.name);
	    break;
	case BADTMP:
	    printf("Error opening temporary file.\n");
	    syslog(LOG_ERR, "Error opening temporary file.\n");
	    break;
	case NOPASSWDFILE:
	    printf("Passwd file doesn't exist.\n");
	    syslog(LOG_ERR, "File %s doesn't exist.\n", fileinfo.name_ptr);
	    break;
	case BADUSERNAME:
	    printf("Bad username.\n");
	    syslog(LOG_ERR, "Bad username.\n");
	    break;
	case BADOPERATION:
	    printf("Tried to perform unknown operation.\n");
	    syslog(LOG_ERR, "Attempted unknown operation.\n");
	    break;
	case BADPATH:
	    printf("Bad path after \"%s\".\n", basename);
	    syslog(LOG_ERR, "Bad path after \"%s\".\n", basename);
	    break;
	case FILEPERMERR:
	    printf("File system permissions don't allow attempted "
		    "operation.\n");
	    syslog(LOG_ERR, "File system permissions don't allow attempted "
		    "operation (%s).\n", userinfo.name);
	    break;
	case NORANDOM:
	    printf("Couldn't open random source.\n");
	    syslog(LOG_ERR, "Couldn't open random source.\n");
	    break;
	default:
	    printf("Program exited with unknown error code %d.\n", err);
	    syslog(LOG_ERR,"Program exited with unknown error code %d.\n", err);
	    break;
    }
    puts("</p></body></html>");

    /* Free any CGI variables */
    if ((cgi_vars = cgiGetVariables(cgi)) != NULL)
	cgiFreeList(cgi_vars);
#else
    openlog(basename, 0, LOG_USER);
    /* switch statement, print message depending on value of err */
    switch (err) {
	case BADHASHTYPE:
	    fprintf(stderr, "Bad hashing algorithm (must be 'd' or 'm').\n");
	    syslog(LOG_ERR, "Bad hashing algorithm (must be 'd' or 'm').\n");
	    break;
	case MINPASSWORDTOOBIG:
	    fprintf(stderr, "Specified minimum password length is too long.\n");
	    syslog(LOG_ERR, "Specified minimum password length is too long.\n");
	    break;
	case MINUSERTOOBIG:
	    fprintf(stderr, "Specified minimum username length is too long.\n");
	    syslog(LOG_ERR, "Specified minimum username length is too long.\n");
	    break;
	case BADCALL:
	    fprintf(stderr, "Unknown option.\n");
	    syslog(LOG_ERR, "Unknown option.\n");
	    break;
	case NOCONFFILE:
	    fprintf(stderr, "Error opening config file.\n");
	    syslog(LOG_ERR, "Error opening config file.\n");
	    break;
	case BADCONFFILE:
	    fprintf(stderr, "Syntax error in config file.\n");
	    syslog(LOG_ERR, "Syntax error in config file.\n");
	    return;
	case USERNOTFOUND:
	    fprintf(stderr, "User %s does not exist in %s.\n", 
		    userinfo.name, fileinfo.name);
	    syslog(LOG_ERR, "User %s does not exist in %s.\n", 
		    userinfo.name, fileinfo.name);
	    break;
	case USEREXISTS:
	    fprintf(stderr, "User %s already exists in %s.\n",
		    userinfo.name, fileinfo.name);
	    syslog(LOG_ERR, "User %s already exists in %s.\n",
		    userinfo.name, fileinfo.name);
	    break;
	case NOAUTHENTICATE:
	    fprintf(stderr, "Permission to change %s denied.\n", fileinfo.name);
	    syslog(LOG_ERR, "Permission to change %s denied (%s).\n", 
		    fileinfo.name, userinfo.name);
	    break;
	case BADPASSWORD:
	    fprintf(stderr, "Bad password or passwords don't match.\n");
	    syslog(LOG_ERR, "Bad password or passwords don't match (%s).\n",
		    userinfo.name);
	    break;
	case BADLOCK:
	    fprintf(stderr, "Could not lock %s.\n", fileinfo.name);
	    syslog(LOG_ERR, "Could not lock %s.\n", fileinfo.name);
	    return;
	case BADTMP:
	    fprintf(stderr, "Error opening temporary file.\n");
	    syslog(LOG_ERR, "Error opening temporary file.\n");
	    return;
	case NOPASSWDFILE:
	    if (fileinfo.name_ptr == NULL) {
		fprintf(stderr, "Password file not specified.\n");
		syslog(LOG_ERR, "Password file not specified.\n");
	    } else {
		fprintf(stderr, "File %s doesn't exist.\n", fileinfo.name_ptr);
		syslog(LOG_ERR, "File %s doesn't exist.\n", fileinfo.name_ptr);
	    }
	    break;
	case BADUSERNAME:
	    fprintf(stderr, "Bad username.\n");
	    syslog(LOG_ERR, "Bad username.\n");
	    break;
	case BADOPERATION:
	    fprintf(stderr, "Attempted unknown operation.\n");
	    syslog(LOG_ERR, "Attempted unknown operation.\n");
	    break;
	case FILEPERMERR:
	    fprintf(stderr, "File system permissions don't allow "
		    "attempted operation.\n");
	    syslog(LOG_ERR, "File system permissions don't allow "
		    "attempted operation (%s).\n", userinfo.name);
	    break;
	case NORANDOM:
	    fprintf(stderr, "Couldn't open random source.\n");
	    syslog(LOG_ERR, "Couldn't open random source.\n");
	    return;
	case GETHELP:
	    break;
	case SHOWVERSION:
	    fprintf(stderr, "chpwdfile version %s\n", version);
	    return;
	default:
	    fprintf(stderr, "Program exited with unknown error "
		    "code %d.\n", err);
	    syslog(LOG_ERR, "Program exited with unknown error "
		    "code %d.\n", err);
	    return;
    }

/*  Modified 10/07/2003 by Mark Valiukas because original version
 *  chokes GCC 3.3 - presumably because 3.3 is more pedantic about
 *  whitespace than 2.95 is
 */
    fprintf(stderr, "\nUsage for %s:\n", basename);
    fprintf(stderr, "\n\t\t%s [options] [username] \n", basename);
    fprintf(stderr, "\n  Options:");
    fprintf(stderr, "\n  -a  add the user             }");
    fprintf(stderr, "\n  -d  delete the user          } Default is -%c", DEFAULTOPERATION);
    fprintf(stderr, "\n  -m  modify the user's entry  }");
    fprintf(stderr, "\n  -c  <filename>");
    fprintf(stderr, "  specify a config file [default=%s]", DEFAULTCONFFILE);
    fprintf(stderr, "\n  -f  <filename>  specify the file ");
    fprintf(stderr, "to modify (must already exist)");
    fprintf(stderr, "\n  -p  <num>");
    fprintf(stderr, "       set minimum password length [default=%d]", DEFAULTMINPASSWD);
    fprintf(stderr, "\n  -u  <num>       set minimum username length [default=%d]", DEFAULTMINPASSWD);
    fprintf(stderr, "\n  -t  <d|m>       set hash type; 'd'==DES, 'm'=MD5 [default=%c]", DEFAULTHASH);
    fprintf(stderr, "\n  -s  read password directly from stdin (useful for scripts)");
    fprintf(stderr, "\n  -h  show usage information");
    fprintf(stderr, "\n  -v  show version number\n\n");


#endif /* CGI */

    closelog();
    return;
}


int main(int argc, char *argv[])
{
    char *basename_ptr;
    char conffile[]=DEFAULTCONFFILE;	/* configuration file */
    char *conffile_ptr=&conffile[0];
    FILE *confstream;
    int err = 0;	/* return code; 0 == no errors */
    int minuser, minpasswd;	/* minimum username & passwd lengths */
    char *username_ptr = NULL;
#ifdef CGI
    char *path_info = NULL;	/* CGI PATH_INFO */
    char **cgi_vars;	/* list of CGI variables */
#else
    int basename_len, counter=0;
    char options[] = "ac:df:hmp:t:u:vs";	/* legal command-line options */
    int opt;	/* options returned by getopt() */
#endif /* CGI */
    int operation;	/* 'a'==add, 'd'==delete, 'm'==modify */
    int authenticate = 1;	/* must we?; 0 == no */
    char logentry[MAXLINELENGTH];
    int minpasswd_opt = -1, minuser_opt = -1, hash_type_opt = -1;

    /* Initialise some variables */
    basename_ptr = argv[0];
    operation = -1;	/* undefined at this point */
    fileinfo.hash_type = DEFAULTHASH;
    minuser = DEFAULTMINUSER;
    minpasswd = DEFAULTMINPASSWD;
    fileinfo.name_ptr = NULL;

#ifdef CGI
    /* Initialise CGI stuff */
    cgiDebug(0, 0);	/* off */
    cgi = cgiInit();

    /* See how we were called and respond accordingly. */
    if ((path_info = getenv("PATH_INFO")) == NULL) {
	err = BADCALL;
	print_usage(basename_ptr, err);
	return(err);
    }
    if (strcmp(path_info, "/add") == 0) { /* Add a user */
	operation = 'a';
    } else if (strcmp(path_info, "/modify") == 0) {	/* Modify entry */
	operation = 'm';
    } else if (strcmp(path_info, "/delete") == 0) {	/* Delete a user */
	operation = 'd';
    } else {	/* Error: the path wasn't valid */
	err = BADPATH;
    }
    if (err) {
	print_usage(basename_ptr, err);
	return(err);
    }
#else
    /* See how we were called and set `operation' */
    /* This behaviour is overridden by command line switches */
    basename_len = (int)strlen(basename_ptr);
    basename_ptr += basename_len;
    while (*basename_ptr != '/' && counter < basename_len+1) {
	basename_ptr--;
	counter++;
    }
    basename_ptr++;
    if (strcmp(basename_ptr, "pwadd") == 0) {
	operation = 'a';
    } else if (strcmp(basename_ptr, "pwdel") == 0) {
	operation = 'd';
    } else if (strcmp(basename_ptr, "pwmod") == 0) {
	operation = 'm';
    } else {
	operation = DEFAULTOPERATION;
    }

    /* Parse command-line options */
    while (((opt=getopt(argc, argv, options)) != -1) && (err == 0)) {
	switch (opt) {
	    case 'a':	/* add an entry */
		operation = opt;
		break;
	    case 'c':	/* location of config file */
		if (getuid() == 0) {
		    strncpy(conffile, optarg, sizeof(conffile));
		    conffile_ptr = optarg;
		}
		break;
	    case 'd':	/* delete an entry */
		operation = opt;
		break;
	    case 'f':	/* location of passwd file */
		strncpy(fileinfo.name, optarg, sizeof(fileinfo.name));
		fileinfo.name_ptr = optarg;
		break;
	    case 'h':	/* help */
		err = GETHELP;
		break;
	    case 'm':	/* modify an entry */
		operation = opt;
		break;
	    case 'p':	/* minimum password length */
		minpasswd_opt = atoi(optarg);
		if (minpasswd_opt > MAXPASSWORDLENGTH)
		    err = MINPASSWORDTOOBIG;
		break;
	    case 't':	/* hash type */
		if (*optarg == 'd' || *optarg == 'm') {
		    hash_type_opt = *optarg;
		} else {
		    err = BADHASHTYPE;
		}
		break;
            case 's':   /* read password from stdin */
                passwd_from_stdin = 1;
                break;
	    case 'u':	/* minimum username length */
		minuser_opt = atoi(optarg);
		if (minuser_opt > MAXUSERLENGTH)
		    err = MINUSERTOOBIG;
		break;
	    case 'v':	/* show version */
		err = SHOWVERSION;
		break;
	    case '?':	/* unknown option */
	    default:
		err = BADCALL;
	}
	if (err) {	/* exit if error */
	    print_usage(basename_ptr, err);
	    return(err);
	}
    }
#endif /* CGI */

    /* Parse config file; command-line options must override these;
     * options based on name of password file. */
    if ((confstream = fopen(conffile_ptr, "r")) != NULL) {
	/* Read config file */
	err = read_config(confstream, &minpasswd, &minuser);
	fclose(confstream);
    } else {
#ifdef CGI
	err = NOCONFFILE;
#else	
	/* Don't complain if default doesn't exist */
	if (strcmp(conffile_ptr, DEFAULTCONFFILE) != 0) {
	    err = NOCONFFILE;
	}
#endif /* CGI */
    }
    if (err) {
	print_usage(basename_ptr, err);
	return(err);
    }

    /* Check that password file was specified and exists; stat() it
     * if things seem OK */
    if (fileinfo.name_ptr == NULL || 
	    stat(fileinfo.name_ptr, &fileinfo.pwdstat) == -1) {
	err = NOPASSWDFILE;
	print_usage(basename_ptr, err);
	return(err);
    }

    /* Check that a user name was specified and is long enough */
#ifdef CGI
    username_ptr = cgiGetValue(cgi, CGIUSERNAME);
#else
    username_ptr = argv[optind];
    if (username_ptr == NULL)
	username_ptr = getenv("USER");
#endif /* CGI */
    if (username_ptr == NULL || (int)strlen(username_ptr) < minuser) {
	err = BADUSERNAME;
	print_usage(basename_ptr, err);
	return(err);
    } else {
	strncpy(userinfo.name, username_ptr, sizeof(userinfo.name)-1);
	userinfo.name[sizeof(userinfo.name)-1] = '\0'; /* Just in case */
    }

#ifndef CGI
    /* Check if executing user has group ownership of passwd file else
     * we will ask for supervisor password or specified user's password.
     * Root never needs to authenticate. */
    if (getuid() == 0) {
	authenticate = 0;
    } else {
	authenticate = check_ownership(fileinfo.pwdstat.st_gid);
    }
#endif

    /* Trusted users may set some options on the command line */
    if (!authenticate) {
	if (hash_type_opt > 0)
	    fileinfo.hash_type = hash_type_opt;
	if (minuser_opt > 0)
	    minuser = minuser_opt;
	if (minpasswd_opt > 0)
	    minpasswd = minpasswd_opt;
    }

    /* Set some defaults if nothing else was specified */
    /* if (fileinfo.hash_type < 0)
	fileinfo.hash_type = DEFAULTHASH;
    if (minuser < 0)
	minuser = DEFAULTMINUSER;
    if (minpasswd < 0)
	minpasswd = DEFAULTMINPASSWD; */

    /* Add, delete or modify entry as desired */
    switch (operation) {
	case 'a':	/* add */
	    /* Prepare for some randomness */
	    random_source = fopen(RANDOMSOURCE, "r");
	    if (random_source == NULL) {
		print_usage(basename_ptr, NORANDOM);
		return(NORANDOM);
	    }
	    err = add_user(minpasswd, authenticate);
	    fclose(random_source);
	    snprintf(logentry, sizeof(logentry), 
		    "Added entry for %s to %s\n", userinfo.name, fileinfo.name);
	    break;
	case 'd':	/* delete */
	    err = delete_user(authenticate);
	    snprintf(logentry, sizeof(logentry), 
		    "Deleted entry for %s from %s\n", 
		    userinfo.name, fileinfo.name);
	    break;
	case 'm':	/* modify */
	    /* Prepare for some randomness */
	    random_source = fopen(RANDOMSOURCE, "r");
	    if (random_source == NULL) {
		print_usage(basename_ptr, NORANDOM);
		return(NORANDOM);
	    }
	    err = modify_user(minpasswd, authenticate);
	    fclose(random_source);
	    snprintf(logentry, sizeof(logentry), 
		    "Modified entry for %s in %s\n", 
		    userinfo.name, fileinfo.name);
	    break;
	default:	/* should never get here */
	    err = BADOPERATION;
    }
    if (err) {
	print_usage(basename_ptr, err);
	/* Remove the temp file if it was made */
	close(fileinfo.tmpfd);
	unlink(fileinfo.tmpname);
	return(err);
    }

    /* We should only have got here if we succeeded */

#ifdef CGI
    /* Free any CGI variables */
    if ((cgi_vars = cgiGetVariables(cgi)) != NULL) 
	cgiFreeList(cgi_vars);
#endif /* CGI */

    /* Write to syslog */
    openlog(basename_ptr, 0, LOG_USER);
    syslog(LOG_NOTICE, "%s", logentry);
    closelog();

#ifdef CGI
    /* Tell the world we succeeded */
    cgiHeader();
    puts("<html><head><title>Success</title></head>");
    puts("<body><h1>Modification successful</h1>");
    puts("<p>Your change has been accepted.</body></html>");
#endif

    return(err);
}


