/* $Id: util.c 1.9 Sun, 01 Apr 2001 23:10:11 +0200 warwick $ */
/* Some generic routines for chpwdfile.  None of these use the global vars. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>

#ifdef __GLIBC__
#include <crypt.h>
#endif

extern int read_new_passwds(char *newpasswd1, char *newpasswd2, int maxlength);

/*
 * Read in and encrypt password of max length 'maxlength' and minimum
 * length 'minlength'.  Return 0 on success.
 * (Actual reading of password is done by a separate function.)
 */
int get_passwd(char *passwd, int maxlength, int minlength, int hash_type,
	FILE *random_source)
{
    int counter, random_number;
    char input_passwd[maxlength], confirm_passwd[maxlength];
    char possible_salt[]="./0123456789"	/* 64 possibilities */
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
    char salt[12], *salt_ptr=salt;
    int salt_length;
    char *crypted_ptr;

    /* Read password twice and check for match */
    if (read_new_passwds(input_passwd, confirm_passwd, maxlength) == 1)
	return(1);
    if (strncmp(input_passwd, confirm_passwd, maxlength) != 0) {
	return(1);
    }

    /* Check that password is long enough */
    if ((int)strlen(input_passwd) < minlength) {
	return(1);
    }

    /* Generate the salt depending on the hash type */
    if (hash_type == 'm') {	/* MD5 algorithm */
	strncpy(salt, "$1$", 3);
	salt_ptr += 3;
	salt_length = 8;	/* 11 less the 3 already set */
    } else {	/* DES algorithm */
	salt_length = 2;
    }
    srand((unsigned int)fgetc(random_source));
    for (counter = 0; counter < salt_length; counter++) {
	random_number = rand();
	srand((unsigned int)fgetc(random_source));
	random_number %= 64;
	*salt_ptr = possible_salt[random_number];
	salt_ptr++;
    }
    *salt_ptr = '\0';

    /* Actually encrypt the password */
    crypted_ptr = crypt(input_passwd, salt);
    while (*crypted_ptr != '\0') {
	*passwd = *crypted_ptr;
	passwd++;
	crypted_ptr++;
    }
    *passwd = '\0';

    return(0);
}

/*
 * Make temporary file and return file pointer and descriptor 
 * in calling list.  Return 0 on success, else 1.
 */
int open_tmpfile(char *pwdfile, char *tmpfile, int *tmpfd, FILE **tmpstream)
{
    int counter;
    char *tmpfile_ptr;	/* points to the start of the tmp file name */

    tmpfile_ptr = tmpfile;

    /* First part of tmpfile == pwdfile */
    while (*pwdfile != '\0') {
	*tmpfile = *pwdfile;
	tmpfile++;
	pwdfile++;
    }

    /* Make unique temp filename and open the tmp file */
    for (counter = 0; counter < 6; counter++) {
	*tmpfile = 'X';
	tmpfile++;
    }
    *tmpfile = '\0';

    *tmpfd = mkstemp(tmpfile_ptr);
    chmod(tmpfile, 00600);

    if ((*tmpstream = fdopen(*tmpfd, "a")) == NULL) {
	return(1);
    }

    return(0);
}

/*
 * Try to lock file; return 0 if successful, 1 if not.
 */
int lock_pwdfile(int fd, int maxtries)
{
    int counter;
    off_t len=0;	/* amount of file to lock; 0 == whole file */

    for (counter = 0; counter < maxtries; counter++) {
	if (lockf(fd, F_TLOCK, len) == 0)
	    return 0;
	sleep(1);	/* wait a while and try again */
    }

    return 1;
}

/*
 * Remove whitespace and characters after '#' from string.
 */
void squash_line(char *line_ptr, char *linecp_ptr)
{
    while (*line_ptr != '\0' && *line_ptr != '#') {
	if (!isspace(*line_ptr)) {
	    *linecp_ptr = *line_ptr;
	    linecp_ptr++;
	}
	line_ptr++;
    }
    *linecp_ptr = '\0';
}

/*
 * Check if the authenticating user has group ownership of the 
 * file.  Returns 0 if this is true.
 */
int check_ownership(gid_t st_gid)
{
    int ngroups;	/* number of groups user belongs to */
    gid_t *groups_ptr;	/* groups user belongs to */
    int notmember=1;	/* 0 == yes, he belong to a relevant group */
    int i;		/* counter */
    gid_t *mem_start;

    ngroups = getgroups(0, NULL);
    groups_ptr = (gid_t *)malloc(sizeof(gid_t)*ngroups);
    mem_start = groups_ptr;
    getgroups(ngroups, groups_ptr);
    for (i = 0; i < ngroups && notmember; i++) {
	if (*groups_ptr == st_gid) 
	    notmember = 0;	/* no authentication needed */
	groups_ptr++;
    }
    free((void *)mem_start);

    return(notmember);
}

/*
 * Authenticate off a given username:passwd pair.  Return 1 on success,
 * 0 otherwise.
 */
int authenticate_user(char *user, char *passwd, char *read_ptr, 
	int maxpasswordlength)
{
    char salt[12];
    int passwd_length;
    char read_passwd[maxpasswordlength];

    /* Refuse to authenticate against an empty password */
    if ((int)strlen(passwd) == 0)
	return(0);

    /* Extract the salt, depending on whether MD5 or DES */
    if (strncmp(passwd, "$1$", 3) == 0) {
	strncpy(salt, passwd, 11);
	passwd_length = 34;
	salt[11] = '\0';
    } else {
	strncpy(salt, passwd, 2);
	passwd_length = 17;
	salt[2] = '\0';
    }

    /* Encrypt the plain text password */
    read_ptr = crypt(read_ptr, salt);
    strncpy(read_passwd, read_ptr, passwd_length);

    /* Check if it matches the user's password */
    if (strncmp(passwd, read_passwd, passwd_length) == 0) {
	return(1);	/* they match */
    } else {
	return(0);	/* they don't */
    }
}
